import os
import unittest
from enum import Enum, unique, auto
import datetime

import pandas as pd
import numpy as np

from helpsk import pandas as hp
from helpsk import validation as hv
from helpsk.utility import redirect_stdout_to_file
from helpsk.validation import iterables_are_equal, assert_dataframes_match
from tests.helpers import get_data_credit, get_test_path, clean_formatted_dataframe


@unique
class TestEnum(Enum):
    VALUE_A = auto()
    VALUE_B = auto()


# noinspection PyMethodMayBeStatic
class TestPandas(unittest.TestCase):

    def helper_test_summary(self, file_name, summary_results):
        if os.path.isfile(file_name):
            os.remove(file_name)
        with redirect_stdout_to_file(file_name):
            hp.print_dataframe(summary_results)
        self.assertTrue(os.path.isfile(file_name))

    @classmethod
    def setUpClass(cls):
        cls.credit_data = get_data_credit()
        sample_data = pd.DataFrame({'col_a': [np.nan, 2, 3, 4],
                                    'col_b': [np.nan, 'b', 'd', 'd'],
                                    'col_c': pd.date_range('2021-01-01', '2021-01-04'),
                                    'col_d': [None, np.nan,
                                              datetime.date(2021, 4, 2), datetime.date(2021, 4, 2)],
                                    'col_e': np.nan,
                                    'col_f': [1.0, 2.0, 3.0, 4.0],
                                    'col_h': [np.nan, TestEnum.VALUE_A, TestEnum.VALUE_B, TestEnum.VALUE_B],
                                    'col_i': [None, np.nan, datetime.datetime(2021, 4, 2, 0, 0, 0),
                                              datetime.datetime(2021, 4, 2, 0, 0, 0)],
                                    'col_j': [False, False, True, False],
                                    'col_k': [None, None, None, None],
                                    'col_l': [np.nan, np.nan, np.nan, np.nan]
                                    })
        sample_data.loc[0, 'col_c'] = np.nan
        sample_data['col_g'] = sample_data['col_b'].astype('category')
        cls.sample_data = sample_data

    def test_is_series_numeric(self):
        actual = self.sample_data.apply(hp.is_series_numeric)
        expected = {
            'col_a': True,
            'col_b': False,
            'col_c': False,
            'col_d': False,
            'col_e': True,
            'col_f': True,
            'col_h': False,
            'col_i': False,
            'col_j': False,
            'col_k': False,
            'col_l': True,
            'col_g': False
        }
        self.assertEqual(expected, actual.to_dict())

    def test_is_series_bool(self):
        self.assertTrue(hp.is_series_bool(series=pd.Series([True])))
        self.assertTrue(hp.is_series_bool(series=pd.Series([False])))
        self.assertTrue(hp.is_series_bool(series=pd.Series([True, False])))
        self.assertTrue(hp.is_series_bool(series=pd.Series([True, False, None])))
        self.assertTrue(hp.is_series_bool(series=pd.Series([True, False, np.nan])))
        self.assertTrue(hp.is_series_bool(series=pd.Series([None, True, False, None])))
        self.assertTrue(hp.is_series_bool(series=pd.Series([np.nan, True, False, np.nan])))

        self.assertFalse(hp.is_series_bool(series=pd.Series([], dtype=object)))
        self.assertFalse(hp.is_series_bool(series=pd.Series(['Whatever'])))
        self.assertFalse(hp.is_series_bool(series=pd.Series([True, False, 'Whatever'])))
        self.assertFalse(hp.is_series_bool(series=pd.Series([True, False, TestEnum.VALUE_A])))
        self.assertFalse(hp.is_series_bool(series=pd.Series([None])))
        self.assertFalse(hp.is_series_bool(series=pd.Series([np.nan])))

    def test_is_series_date(self):
        self.assertFalse(hp.is_series_date(self.sample_data['col_a']))
        self.assertFalse(hp.is_series_date(self.sample_data['col_b']))
        self.assertTrue(hp.is_series_date(self.sample_data['col_c']))
        self.assertTrue(hp.is_series_date(self.sample_data['col_d']))
        self.assertFalse(hp.is_series_date(self.sample_data['col_e']))
        self.assertFalse(hp.is_series_date(self.sample_data['col_f']))
        self.assertFalse(hp.is_series_date(self.sample_data['col_g']))
        self.assertFalse(hp.is_series_date(self.sample_data['col_h']))
        self.assertTrue(hp.is_series_date(self.sample_data['col_i']))
        self.assertFalse(hp.is_series_date(pd.Series(dtype=np.float64)))

    def test_is_series_date__non_numeric_index(self):
        data = self.sample_data.copy()
        data.index = ['A', 'B', 'C', 'D']
        self.assertEqual(hp.get_date_columns(data), ['col_c', 'col_d', 'col_i'])

    def test_is_series_string(self):
        actual = self.sample_data.apply(hp.is_series_string)
        expected = {
            'col_a': False,
            'col_b': True,
            'col_c': False,
            'col_d': False,
            'col_e': False,
            'col_f': False,
            'col_h': False,
            'col_i': False,
            'col_j': False,
            'col_k': False,
            'col_l': False,
            'col_g': False
        }
        self.assertEqual(expected, actual.to_dict())
        self.assertEqual(hp.get_string_columns(self.sample_data), ['col_b'])

        data = self.sample_data.copy()
        data.index = ['A', 'B', 'C', 'D']
        self.assertEqual(hp.get_string_columns(data), ['col_b'])

    def test_is_series_categorical(self):
        actual = self.sample_data.apply(hp.is_series_categorical)
        expected = {
            'col_a': False,
            'col_b': False,
            'col_c': False,
            'col_d': False,
            'col_e': False,
            'col_f': False,
            'col_h': False,
            'col_i': False,
            'col_j': False,
            'col_k': False,
            'col_l': False,
            'col_g': True
        }
        self.assertEqual(expected, actual.to_dict())
        self.assertEqual(hp.get_categorical_columns(self.sample_data), ['col_g'])

    def test_replace_all_bools_with_strings(self):
        results = hp.replace_all_bools_with_strings(series=pd.Series([True]))
        self.assertEqual(list(results), ['True'])
        results = hp.replace_all_bools_with_strings(series=pd.Series([False]))
        self.assertEqual(list(results), ['False'])
        results = hp.replace_all_bools_with_strings(series=pd.Series([True, False]))
        self.assertEqual(list(results), ['True', 'False'])
        results = hp.replace_all_bools_with_strings(series=pd.Series([True, False]).astype('category'))
        self.assertEqual(list(results), ['True', 'False'])
        results = hp.replace_all_bools_with_strings(series=pd.Series([True, False, None]))
        self.assertEqual(list(results), ['True', 'False', None])
        results = hp.replace_all_bools_with_strings(series=pd.Series([True, False, np.nan]))
        self.assertEqual(list(results), ['True', 'False', np.nan])
        results = hp.replace_all_bools_with_strings(series=pd.Series([None, True, False, None]))
        self.assertEqual(list(results), [None, 'True', 'False', None])
        results = hp.replace_all_bools_with_strings(series=pd.Series([np.nan, True, False, np.nan]))
        self.assertEqual(list(results), [np.nan, 'True', 'False', np.nan])

        results = hp.replace_all_bools_with_strings(series=pd.Series(['True']))
        self.assertEqual(list(results), ['True'])
        results = hp.replace_all_bools_with_strings(series=pd.Series(['False']))
        self.assertEqual(list(results), ['False'])
        results = hp.replace_all_bools_with_strings(series=pd.Series(['True', 'False']))
        self.assertEqual(list(results), ['True', 'False'])
        results = hp.replace_all_bools_with_strings(series=pd.Series(['True', 'False']).astype('category'))
        self.assertEqual(list(results), ['True', 'False'])
        results = hp.replace_all_bools_with_strings(series=pd.Series(['True', 'False', None]))
        self.assertEqual(list(results), ['True', 'False', None])
        results = hp.replace_all_bools_with_strings(series=pd.Series(['True', 'False', np.nan]))
        self.assertEqual(list(results), ['True', 'False', np.nan])
        results = hp.replace_all_bools_with_strings(series=pd.Series([None, 'True', 'False', None]))
        self.assertEqual(list(results), [None, 'True', 'False', None])
        results = hp.replace_all_bools_with_strings(series=pd.Series([np.nan, 'True', 'False', np.nan]))
        self.assertEqual(list(results), [np.nan, 'True', 'False', np.nan])

        results = hp.replace_all_bools_with_strings(series=pd.Series([], dtype=object))
        self.assertEqual(list(results), [])
        results = hp.replace_all_bools_with_strings(series=pd.Series(['Whatever']))
        self.assertEqual(list(results), ['Whatever'])
        results = hp.replace_all_bools_with_strings(series=pd.Series([True, False, 'Whatever']))
        self.assertEqual(list(results), ['True', 'False', 'Whatever'])
        results = hp.replace_all_bools_with_strings(series=pd.Series([True, False, TestEnum.VALUE_A]))
        self.assertEqual(list(results), ['True', 'False', TestEnum.VALUE_A])
        results = hp.replace_all_bools_with_strings(series=pd.Series([None]))
        self.assertEqual(list(results), [None])
        results = hp.replace_all_bools_with_strings(series=pd.Series([np.nan]))
        self.assertEqual(len(results), 1)

    def test_fill_na(self):
        # test non-categorical
        self.assertEqual(hp.fill_na(pd.Series([], dtype=object)).values.tolist(), [])
        self.assertEqual(hp.fill_na(pd.Series([None])).values.tolist(), ['<Missing>'])
        self.assertEqual(hp.fill_na(pd.Series([np.nan])).values.tolist(), ['<Missing>'])

        results = hp.fill_na(pd.Series([True, False, np.nan]))
        self.assertEqual(results.tolist(), [True, False, '<Missing>'])
        self.assertIsInstance(results[0], bool)

        # this fails if not converted to object
        results = hp.fill_na(pd.Series([True, False, np.nan]).astype('boolean'))
        self.assertEqual(results.tolist(), [True, False, '<Missing>'])
        self.assertIsInstance(results[0], bool)

        results = hp.fill_na(pd.Series([0, 1, np.nan]))
        self.assertEqual(results.tolist(), [0, 1, '<Missing>'])
        self.assertIsInstance(results[0], float)

        # this fails if not converted to object
        results = hp.fill_na(pd.Series([0, 1, np.nan]).astype('float'))
        self.assertEqual(results.tolist(), [0, 1, '<Missing>'])
        self.assertIsInstance(results[0], float)

        results = hp.fill_na(pd.Series(['A', 'B', np.nan]))
        self.assertEqual(results.tolist(), ['A', 'B', '<Missing>'])
        self.assertIsInstance(results[0], str)

        self.assertEqual(hp.fill_na(pd.Series(['A', 'B'])).values.tolist(), ['A', 'B'])

        # test categorical
        results = hp.fill_na(pd.Series([], dtype=object).astype('category'))
        self.assertEqual(results.tolist(), [])
        self.assertTrue(hp.is_series_categorical(results))
        self.assertTrue('<Missing>' in results.cat.categories)

        results = hp.fill_na(pd.Series([None]).astype('category'))
        self.assertEqual(results.tolist(), ['<Missing>'])
        self.assertTrue(hp.is_series_categorical(results))
        self.assertTrue('<Missing>' in results.cat.categories)

        results = hp.fill_na(pd.Series([np.nan]).astype('category'))
        self.assertEqual(results.tolist(), ['<Missing>'])
        self.assertTrue(hp.is_series_categorical(results))
        self.assertTrue('<Missing>' in results.cat.categories)

        results = hp.fill_na(pd.Series([True, False, np.nan]).astype('category'))
        self.assertEqual(results.tolist(), [True, False, '<Missing>'])
        self.assertTrue(hp.is_series_categorical(results))
        self.assertTrue('<Missing>' in results.cat.categories)

        results = hp.fill_na(pd.Series([0, 1, np.nan]).astype('category'))
        self.assertEqual(results.tolist(), [0, 1, '<Missing>'])
        self.assertTrue(hp.is_series_categorical(results))
        self.assertTrue('<Missing>' in results.cat.categories)

        results = hp.fill_na(pd.Series(['A', 'B', '<Missing>', np.nan]).astype('category'))
        self.assertEqual(results.tolist(), ['A', 'B', '<Missing>', '<Missing>'])
        self.assertTrue(hp.is_series_categorical(results))
        self.assertTrue('<Missing>' in results.cat.categories)

        results = hp.fill_na(pd.Series(['A', 'B', np.nan]).astype('category'))
        self.assertEqual(results.tolist(), ['A', 'B', '<Missing>'])
        self.assertTrue(hp.is_series_categorical(results))
        self.assertTrue('<Missing>' in results.cat.categories)

        results = hp.fill_na(pd.Series(['A', 'B']).astype('category'))
        self.assertEqual(results.tolist(), ['A', 'B'])
        self.assertTrue(hp.is_series_categorical(results))
        self.assertTrue('<Missing>' in results.cat.categories)

    def test_get_numeric_columns(self):
        self.assertEqual(hp.get_numeric_columns(self.sample_data), ['col_a', 'col_e', 'col_f', 'col_l'])
        self.assertEqual(hp.get_numeric_columns(self.sample_data[['col_e']]), ['col_e'])
        self.assertEqual(hp.get_numeric_columns(self.sample_data[['col_d']]), [])
        self.assertEqual(hp.get_numeric_columns(pd.DataFrame()), [])

    def test_get_non_numeric_columns(self):
        non_numeric_columns = hp.get_non_numeric_columns(self.sample_data)
        non_numeric_columns.sort()
        self.assertEqual(non_numeric_columns,
                         ['col_b', 'col_c', 'col_d', 'col_g', 'col_h', 'col_i', 'col_j', 'col_k'])
        self.assertEqual(hp.get_non_numeric_columns(self.sample_data[['col_d']]), ['col_d'])
        self.assertEqual(hp.get_non_numeric_columns(self.sample_data[['col_e']]), [])
        self.assertEqual(hp.get_non_numeric_columns(pd.DataFrame()), [])

    def test_reorder_categories__occurrences(self):
        categorical = self.credit_data['purpose'].copy()
        categorical[0:50] = np.nan
        original_categorical = categorical.copy()

        original_categories = ['new car', 'used car', 'furniture/equipment', 'radio/tv',
                               'domestic appliance', 'repairs', 'education', 'vacation', 'retraining',
                               'business', 'other']

        expected_categories = list(categorical.value_counts(ascending=True, dropna=True).index.values)
        expected_categories_rev = expected_categories.copy()
        expected_categories_rev.reverse()

        # check that the purpose column hasn't changed and that it is not ordered and has original categories
        self.assertFalse(categorical.cat.ordered)
        self.assertEqual(list(categorical.cat.categories), original_categories)

        # test default weights
        # also, we are testing a Categorical object; verify it is categorical, then later test non-categorical
        self.assertEqual(categorical.dtype.name, 'category')
        results = hp.reorder_categories(categorical=categorical, ascending=True, ordered=False)
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertTrue(iterables_are_equal(results, original_categorical))
        self.assertEqual(list(results.categories), expected_categories)
        self.assertFalse(results.ordered)

        # check no side effects
        results[0] = 'new car'
        self.assertEqual(list(original_categorical.cat.categories), original_categories)
        self.assertTrue(pd.isna(original_categorical[0]))

        # test ascending=False & ordered=True
        results = hp.reorder_categories(categorical=categorical, ascending=False, ordered=True)
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertTrue(iterables_are_equal(results, original_categorical))
        self.assertEqual(list(results.categories), expected_categories_rev)
        self.assertTrue(results.ordered)

        # check no side effects
        results[0] = 'new car'
        self.assertEqual(list(categorical.cat.categories), list(original_categorical.cat.categories))
        self.assertTrue(iterables_are_equal(categorical, original_categorical))
        self.assertTrue(pd.isna(categorical[0]))

        series = pd.Series(['a', 'b', 'b', 'c', 'b', 'c'])
        results = hp.reorder_categories(categorical=series, ascending=False, ordered=True)
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertTrue(iterables_are_equal(results, series))
        self.assertEqual(list(results.categories), ['b', 'c', 'a'])
        self.assertTrue(results.ordered)

    def test_reorder_categories__weights(self):
        categorical = self.credit_data['purpose'].copy()
        categorical[0:50] = np.nan
        original_categorical = categorical.copy()
        weights = self.credit_data['credit_amount'].copy()
        weights[25:75] = np.nan

        original_categories = ['new car', 'used car', 'furniture/equipment', 'radio/tv',
                               'domestic appliance', 'repairs', 'education', 'vacation', 'retraining',
                               'business', 'other']

        expected_categories = ['vacation', 'retraining', 'domestic appliance', 'repairs', 'education',
                               'radio/tv', 'new car', 'furniture/equipment', 'business', 'used car', 'other']
        expected_categories_rev = expected_categories.copy()
        expected_categories_rev.reverse()

        with self.assertRaises(hv.HelpskParamValueError):
            hp.reorder_categories(categorical=categorical[0:-1], weights=weights)

        with self.assertRaises(hv.HelpskParamValueError):
            hp.reorder_categories(categorical=categorical, weights=weights[0:-1])

        # check that the purpose column hasn't changed and that it is not ordered and has original categories
        self.assertFalse(categorical.cat.ordered)
        self.assertEqual(list(categorical.cat.categories), original_categories)

        # test default weights
        # also, we are testing a Categorical object; verify it is categorical, then later test non-categorical
        self.assertEqual(categorical.dtype.name, 'category')
        results = hp.reorder_categories(
            categorical=categorical, weights=weights, ascending=True, ordered=False
        )
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertTrue(iterables_are_equal(results, original_categorical))
        self.assertEqual(list(results.categories), expected_categories)
        self.assertFalse(results.ordered)

        # check no side effects
        results[0] = 'new car'
        self.assertEqual(list(categorical.cat.categories), list(original_categorical.cat.categories))
        self.assertTrue(iterables_are_equal(categorical, original_categorical))
        self.assertTrue(pd.isna(categorical[0]))

        # test ascending=False & ordered=True
        results = hp.reorder_categories(
            categorical=categorical, weights=weights, ascending=False, ordered=True
        )
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertTrue(iterables_are_equal(results, original_categorical))
        self.assertEqual(list(results.categories), expected_categories_rev)
        self.assertTrue(results.ordered)

        # check no side effects
        results[0] = 'new car'
        self.assertEqual(list(categorical.cat.categories), list(original_categorical.cat.categories))
        self.assertTrue(iterables_are_equal(categorical, original_categorical))
        self.assertTrue(pd.isna(categorical[0]))

        series = pd.Series(['a', 'b', 'c'] * 2)
        weights = pd.Series([1, 3, 2] * 2)
        results = hp.reorder_categories(categorical=series, weights=weights, ascending=False, ordered=True)
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertTrue(iterables_are_equal(results, series))
        self.assertEqual(list(results.categories), ['b', 'c', 'a'])
        self.assertTrue(results.ordered)

    def test_top_n_categories__occurrences(self):
        categorical = self.credit_data['purpose'].copy()
        categorical[0:50] = np.nan
        original_categorical = categorical.copy()
        categorical_counts = categorical.value_counts(dropna=False)

        results = hp.top_n_categories(categorical=categorical)
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertEqual(list(results.categories), list(categorical_counts.index.values[0:5]) + ['Other'])
        self.assertEqual(results.value_counts(dropna=False).sum(), 1000)

        result_counts = results.value_counts().head(-1)
        expected_counts = categorical_counts.head(5)
        self.assertEqual(list(result_counts.index), list(expected_counts.index))
        self.assertEqual(list(result_counts.values), list(expected_counts.values))

        # check no side effects
        results[0] = 'new car'
        self.assertEqual(list(categorical.cat.categories), list(original_categorical.cat.categories))
        self.assertTrue(iterables_are_equal(categorical, original_categorical))
        self.assertTrue(pd.isna(categorical[0]))

        # if top_n is set to a higher number than there are categories, then we should get back an object
        # with the same values
        results = hp.top_n_categories(categorical=categorical, top_n=20)
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertTrue(set(results.cat.categories) == set(categorical.cat.categories))  # noqa
        self.assertTrue(iterables_are_equal(results, categorical))

        # check no side effects
        results[0] = 'new car'
        self.assertEqual(list(categorical.cat.categories), list(original_categorical.cat.categories))
        self.assertTrue(iterables_are_equal(categorical, original_categorical))
        self.assertTrue(pd.isna(categorical[0]))

        results = hp.top_n_categories(
            categorical=categorical,
            top_n=3,
            other_category='Some others',
            ordered=True)
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertEqual(list(results.categories),
                         list(categorical_counts.index.values[0:3]) + ['Some others'])
        self.assertEqual(results.value_counts(dropna=False).sum(), 1000)

        result_counts = results.value_counts().head(-1)
        expected_counts = categorical_counts.head(3)
        self.assertEqual(list(result_counts.index), list(expected_counts.index))
        self.assertEqual(list(result_counts.values), list(expected_counts.values))

        # check no side effects
        results[0] = 'new car'
        self.assertEqual(list(categorical.cat.categories), list(original_categorical.cat.categories))
        self.assertTrue(iterables_are_equal(categorical, original_categorical))
        self.assertTrue(pd.isna(categorical[0]))

    def test_top_n_categories__weights(self):
        categorical = self.credit_data['purpose'].copy()
        categorical[0:50] = np.nan
        original_categorical = categorical.copy()
        weights = pd.Series([-1] * len(categorical))
        # ascending True because we are going sum by negative -1 so categories with least amount of
        # occurrences will be in the "top" categories i.e. -1 > -100
        # now we expect our top_n to be the lowest frequency of categories
        categorical_counts = categorical.value_counts(dropna=False, ascending=True)

        with self.assertRaises(hv.HelpskParamValueError):
            hp.top_n_categories(categorical=categorical[0:-1], weights=weights)

        with self.assertRaises(hv.HelpskParamValueError):
            hp.top_n_categories(categorical=categorical, weights=weights[0:-1])

        results = hp.top_n_categories(categorical=categorical, weights=weights, weight_function=np.sum)
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertEqual(list(results.categories), list(categorical_counts.index.values[0:5]) + ['Other'])
        self.assertEqual(results.value_counts(dropna=False).sum(), 1000)

        result_counts = results.value_counts().head(-1)
        expected_counts = categorical_counts.head(5)
        self.assertEqual(list(result_counts.index), list(expected_counts.index))
        self.assertEqual(list(result_counts.values), list(expected_counts.values))

        # check no side effects
        results[0] = 'vacation'
        self.assertEqual(list(categorical.cat.categories), list(original_categorical.cat.categories))
        self.assertTrue(iterables_are_equal(categorical, original_categorical))
        self.assertTrue(pd.isna(categorical[0]))

        # if top_n is set to a higher number than there are categories, then we should get back an object
        # with the same values
        results = hp.top_n_categories(
            categorical=categorical,
            top_n=20,
            weights=weights,
            weight_function=np.sum
        )
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertTrue(set(results.cat.categories) == set(categorical.cat.categories))  # noqa
        self.assertTrue(iterables_are_equal(results, categorical))

        # check no side effects
        results[0] = 'vacation'
        self.assertEqual(list(categorical.cat.categories), list(original_categorical.cat.categories))
        self.assertTrue(iterables_are_equal(categorical, original_categorical))
        self.assertTrue(pd.isna(categorical[0]))

        results = hp.top_n_categories(
            categorical=categorical, top_n=3, other_category='Some others',
            ordered=True, weights=weights, weight_function=np.sum
        )
        self.assertTrue(results is not categorical)  # check that a new object was returned
        self.assertEqual(list(results.categories),
                         list(categorical_counts.index.values[0:3]) + ['Some others'])
        self.assertEqual(results.value_counts(dropna=False).sum(), 1000)

        result_counts = results.value_counts().head(-1)
        expected_counts = categorical_counts.head(3)
        self.assertEqual(list(result_counts.index), list(expected_counts.index))
        self.assertEqual(list(result_counts.values), list(expected_counts.values))

        # check no side effects
        results[0] = 'vacation'
        self.assertEqual(list(categorical.cat.categories), list(original_categorical.cat.categories))
        self.assertTrue(iterables_are_equal(categorical, original_categorical))
        self.assertTrue(pd.isna(categorical[0]))

    def test_numeric_summary(self):
        self.helper_test_summary(
            get_test_path('pandas/test_numeric_summary__credit.txt'),
            hp.numeric_summary(self.credit_data, return_style=False, sort_by_columns=False)
        )

        self.helper_test_summary(
            get_test_path('pandas/test_numeric_summary__credit__sorted.txt'),
            hp.numeric_summary(self.credit_data, return_style=False, sort_by_columns=True)
        )

        self.helper_test_summary(
            get_test_path('pandas/test_numeric_summary__sample.txt'),
            hp.numeric_summary(self.sample_data, return_style=False, sort_by_columns=False)
        )

        self.helper_test_summary(
            get_test_path('pandas/test_numeric_summary__sample__sorted.txt'),
            hp.numeric_summary(self.sample_data, return_style=False, sort_by_columns=True)
        )

    def test_numeric_summary__nan_column(self):
        test_data = self.credit_data.copy()
        test_data['all_missing'] = np.nan
        self.helper_test_summary(
            get_test_path('pandas/test_numeric_summary__credit__all_missing.txt'),
            hp.numeric_summary(test_data, return_style=False)
        )

        self.helper_test_summary(
            get_test_path('pandas/test_numeric_summary__style__credit__all_missing.html'),
            clean_formatted_dataframe(hp.numeric_summary(test_data, return_style=True, sort_by_columns=False).to_html())  # noqa
        )

        self.helper_test_summary(
            get_test_path('pandas/test_numeric_summary__style__credit__all_missing__sorted.html'),  # noqa
            clean_formatted_dataframe(hp.numeric_summary(test_data, return_style=True, sort_by_columns=True).to_html())  # noqa
        )

    def test_numeric_summary_style(self):
        test_data = self.credit_data.copy()
        test_data.loc[0:46, ['duration']] = np.nan
        test_data.loc[10:54, ['credit_amount']] = 0

        self.helper_test_summary(
            get_test_path('pandas/test_numeric_summary__style__credit.html'),
            clean_formatted_dataframe(hp.numeric_summary(test_data, return_style=True, sort_by_columns=False).to_html())  # noqa
        )

        self.helper_test_summary(
            get_test_path('pandas/test_numeric_summary__style__credit__sorted.html'),
            clean_formatted_dataframe(hp.numeric_summary(test_data, return_style=True, sort_by_columns=True).to_html())  # noqa
        )

        self.helper_test_summary(
            get_test_path('pandas/test_numeric_summary__style__sample.html'),
            clean_formatted_dataframe(hp.numeric_summary(self.sample_data, return_style=True).to_html())  # noqa
        )

    def test_non_numeric_summary(self):
        credit_data = self.credit_data.copy()
        credit_data['purpose'] = credit_data['purpose']. \
            replace({'radio/tv': '1111111111222222222233333333334444444444'})

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__credit.txt'),
            hp.non_numeric_summary(credit_data, return_style=False, sort_by_columns=False)
        )

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__credit__sorted.txt'),
            hp.non_numeric_summary(credit_data, return_style=False, sort_by_columns=True)
        )

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__sample.txt'),
            hp.non_numeric_summary(self.sample_data, return_style=False, sort_by_columns=False)
        )

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__sample__sorted.txt'),
            hp.non_numeric_summary(self.sample_data, return_style=False, sort_by_columns=True)
        )

    def test_non_numeric_summary__nan_column(self):
        test_data = self.credit_data.copy()
        test_data['purpose'] = test_data['purpose']. \
            replace({'radio/tv': '1111111111222222222233333333334444444444'})
        test_data['all_missing'] = None
        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__credit__all_missing.txt'),
            hp.non_numeric_summary(test_data, return_style=False, sort_by_columns=False)
        )

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__credit__all_missing__sorted.txt'),
            hp.non_numeric_summary(test_data, return_style=False, sort_by_columns=True)
        )

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__style__credit__all_missing.html'),
            clean_formatted_dataframe(hp.non_numeric_summary(test_data, return_style=True, sort_by_columns=False).to_html())  # noqa
        )

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__style__credit__all_missing__sorted.html'),  # noqa
            clean_formatted_dataframe(hp.non_numeric_summary(test_data, return_style=True, sort_by_columns=True).to_html())  # noqa
        )

    def test_non_numeric_summary__list_column(self):
        test_data = pd.DataFrame({
            'list_column': [['a', 'b'], [1, 2], []],
            'list_column2': [['a', 'b'], [1, 2], np.nan],
        })

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__list_column.txt'),
            hp.non_numeric_summary(test_data, return_style=False, sort_by_columns=False)
        )

    def test_non_numeric_summary2(self):
        test_data = self.credit_data.copy()
        test_data['purpose'] = test_data['purpose']. \
            replace({'radio/tv': '1111111111222222222233333333334444444444'})
        test_data.loc[25:75, ['checking_status']] = np.nan

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__style__credit.html'),
            clean_formatted_dataframe(hp.non_numeric_summary(test_data, return_style=True, sort_by_columns=False).to_html())  # noqa
        )

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__style__credit__sorted.html'),
            clean_formatted_dataframe(hp.non_numeric_summary(test_data, return_style=True, sort_by_columns=True).to_html())  # noqa
        )

        self.helper_test_summary(
            get_test_path('pandas/test_non_numeric_summary__style__sample.html'),
            clean_formatted_dataframe(hp.non_numeric_summary(self.sample_data, return_style=True).to_html())
        )

    def test_convert_integer_series_to_categorical(self):
        # check that function fails if mapping doesn't contains all numbers in data
        with self.assertRaises(hv.HelpskParamValueError):
            mapping = {2: 'a', 3: 'b'}
            hp.convert_integer_series_to_categorical(series=self.sample_data.col_a, mapping=mapping)

        with self.assertRaises(hv.HelpskParamValueError):
            mapping = {2: 'a'}
            hp.convert_integer_series_to_categorical(series=self.sample_data.col_a, mapping=mapping)

        with self.assertRaises(hv.HelpskParamValueError):
            mapping = {}
            hp.convert_integer_series_to_categorical(series=self.sample_data.col_a, mapping=mapping)

        # check that series will work even if it doesn't have all the values in mapping
        mapping = {1: 'value not found', 2: 'a', 3: 'b', 4: 'c'}
        result = hp.convert_integer_series_to_categorical(
            series=self.sample_data.col_a,
            mapping=mapping,
            ordered=False
        )
        self.assertTrue(hv.iterables_are_equal(result.cat.categories, mapping.values()))
        self.assertTrue(hv.iterables_are_equal(list(result.values), [np.nan, 'a', 'b', 'c']))
        # ensure no side effects on list passed in
        self.assertTrue(hv.iterables_are_equal(self.sample_data.col_a, [np.nan, 2, 3, 4]))
        self.assertFalse(result.cat.ordered)

        # same thing but with ordered Categorical
        mapping = {1: 'not found', 2: 'a', 3: 'b', 4: 'c'}
        result = hp.convert_integer_series_to_categorical(
            series=self.sample_data.col_a,
            mapping=mapping,
            ordered=True
        )
        self.assertTrue(hv.iterables_are_equal(result.cat.categories, mapping.values()))
        self.assertTrue(hv.iterables_are_equal(list(result.values), [np.nan, 'a', 'b', 'c']))
        # ensure no side effects on list passed in
        self.assertTrue(hv.iterables_are_equal(self.sample_data.col_a, [np.nan, 2, 3, 4]))
        self.assertTrue(result.cat.ordered)

    def test_value_frequency(self):
        credit_history_order = ['delayed previously',
                                'critical/other existing credit',
                                'existing paid',
                                'no credits/all paid',
                                'all paid']
        credit_history_alphabetical = credit_history_order.copy()
        credit_history_alphabetical.sort()

        test_series_unordered = self.credit_data['credit_history'].copy()
        self.assertFalse(test_series_unordered.cat.ordered)
        test_series_unordered[0:10] = np.nan

        test_series_ordered = test_series_unordered.cat.reorder_categories(credit_history_order, ordered=True)
        self.assertTrue(test_series_ordered.cat.ordered)

        results = hp.value_frequency(test_series_unordered, sort_by_frequency=True)
        expected_value_order = test_series_unordered.value_counts(dropna=False).index.tolist()
        self.assertEqual(results.index.tolist(), expected_value_order)
        self.assertTrue((results['Frequency'] == test_series_unordered.value_counts(dropna=False).values).all())  # noqa
        self.assertTrue((results['Percent'] == test_series_unordered.value_counts(normalize=True, dropna=False).values).all())  # noqa

        # cache the verified results
        cached_results = results

        # test unordered categorical - sort_by_frequency=False
        results = hp.value_frequency(test_series_unordered, sort_by_frequency=False)
        self.assertEqual(results.index.tolist(), credit_history_alphabetical + [np.nan])
        self.assertEqual(
            cached_results.loc[results.index.tolist(), 'Frequency'].tolist(),
            results['Frequency'].tolist()
        )
        self.assertTrue(cached_results.loc[results.index.tolist(), 'Percent'].tolist() == results['Percent'].tolist())  # noqa

        # test ordered categorical - sort_by_frequency=False
        results = hp.value_frequency(test_series_ordered, sort_by_frequency=False)
        self.assertEqual(results.index.tolist(), credit_history_order + [np.nan])
        self.assertTrue(cached_results.loc[results.index.tolist(), 'Frequency'].tolist() == results['Frequency'].tolist())  # noqa
        self.assertTrue(cached_results.loc[results.index.tolist(), 'Percent'].tolist() == results['Percent'].tolist())  # noqa

        # test ordered categorical - sort_by_frequency=True
        results = hp.value_frequency(test_series_ordered, sort_by_frequency=True)
        self.assertEqual(results.index.tolist(), cached_results.index.tolist())
        self.assertTrue(cached_results['Frequency'].tolist() == results['Frequency'].tolist())
        self.assertTrue(cached_results['Percent'].tolist() == results['Percent'].tolist())
        self.assertTrue(hv.dataframes_match([cached_results, results]))

        # test numeric
        test_series_float = self.credit_data['duration'].copy()
        test_series_float[0:10] = np.nan

        results = hp.value_frequency(test_series_float, sort_by_frequency=True)
        expected_value_counts = test_series_float.value_counts(normalize=False, dropna=False)
        self.assertTrue(hv.iterables_are_equal(results.index.values, expected_value_counts.index.values))
        self.assertTrue(hv.iterables_are_equal(results['Frequency'].values, expected_value_counts.values))

        cached_results = results

        results = hp.value_frequency(test_series_float, sort_by_frequency=False)
        expected_indexes = test_series_float.dropna().unique()
        expected_indexes.sort()
        expected_indexes = expected_indexes.tolist() + [np.nan]

        self.assertEqual(results.index.values.tolist()[0:-1], expected_indexes[0:-1])
        self.assertTrue(hv.iterables_are_equal(results.index.values.tolist(), expected_indexes))  # noqa
        self.assertTrue(hv.iterables_are_equal(results['Frequency'].values, cached_results.loc[results.index.values, 'Frequency'].values))  # noqa
        self.assertTrue(hv.iterables_are_equal(results['Percent'].values, cached_results.loc[results.index.values, 'Percent'].values))  # noqa

    def test_value_frequency__missing_category(self):
        # found a bug when doing `value_frequency(series, sort_by_frequency=False)` with a series that had
        # a count of `0` for a category (i.e. category existed but not any values)
        test_series = self.credit_data['checking_status'].copy()
        test_series[test_series == 'no checking'] = np.nan

        results = hp.value_frequency(test_series, sort_by_frequency=True)
        expected_value_counts = test_series.value_counts(normalize=False, dropna=False)

        self.assertTrue(hv.iterables_are_equal(results.index.values, expected_value_counts.index.values))
        self.assertTrue(hv.iterables_are_equal(results['Frequency'].values, expected_value_counts.values))

        cached_results = results

        results = hp.value_frequency(test_series, sort_by_frequency=False)
        expected_indexes = ['0<=X<200', '<0', '>=200', 'no checking', np.nan]

        self.assertEqual(results.index.values.tolist()[0:-1], expected_indexes[0:-1])
        self.assertTrue(hv.iterables_are_equal(results.index.values.tolist(), expected_indexes))  # noqa
        self.assertTrue(hv.iterables_are_equal(results['Frequency'].values, cached_results.loc[results.index.values, 'Frequency'].values))  # noqa
        self.assertTrue(hv.iterables_are_equal(results['Percent'].values, cached_results.loc[results.index.values, 'Percent'].values))  # noqa

    def test_count_groups(self):
        data = self.credit_data.copy()
        data.loc[:, 'target'] = data['target'] == 'good'
        data.loc[0:10, 'target'] = np.nan
        data.loc[:, 'target'] = data['target'].astype('boolean')
        data.loc[9:20, 'checking_status'] = np.nan
        data.loc[19:30, 'credit_amount'] = np.nan

        # change dataset so that there is an entire category missing (e.g. no target == True and checking
        # status `<Missing>`
        indexes_to_blank = (data['target'] == True) & (data['checking_status'] == '0<=X<200')  # noqa
        data.loc[indexes_to_blank, 'checking_status'] = np.nan
        # change dataset so there is an entire category with no sum_by
        indexes_to_blank = (data['target'] == False) & (data['checking_status'] == 'no checking')  # noqa
        data.loc[indexes_to_blank, 'credit_amount'] = np.nan

        original = self.credit_data.copy()
        result = hp.count_groups(
            dataframe=original,
            group_1='target',
            group_2='housing',
            group_sum='credit_amount',
            return_style=False
        )
        assert_dataframes_match([original, self.credit_data])
        file_name = get_test_path('pandas/count_groups__credit.txt')
        with redirect_stdout_to_file(file_name):
            hp.print_dataframe(result)

        self.assertEqual(result[('target', 'target')].dropna().tolist(), ['good', 'bad'])
        self.assertEqual(result[('target', 'Count')].sum(), 1000)
        self.assertEqual(result[('target', 'Count Perc')].sum(), 1)
        self.assertEqual(result[('target', 'Sum')].sum(), self.credit_data['credit_amount'].sum())
        self.assertEqual(result[('target', 'Sum Perc')].sum(), 1)
        self.assertEqual(result[('housing', 'Count')].sum(), 1000)
        self.assertEqual(result[('housing', 'Count Perc')].sum(), 2)
        self.assertEqual(result[('housing', 'Sum')].sum(), self.credit_data['credit_amount'].sum())
        hv.assert_is_close(result[('housing', 'Sum Perc')].sum(), 2)

        def test_count_groups(group_1, group_2=None, group_sum=None, remove_first_level_duplicates=False):
            counts = hp.count_groups(
                dataframe=data,
                group_1=group_1,
                group_2=group_2,
                group_sum=group_sum,
                remove_first_level_duplicates=remove_first_level_duplicates,
                return_style=False
            )

            name = get_test_path(
                f'pandas/count_groups_{group_1}_{group_2}_'
                f'{group_sum}_{remove_first_level_duplicates}.txt'
            )
            with redirect_stdout_to_file(name):
                hp.print_dataframe(counts)
            return counts

        results_1 = test_count_groups(group_1='target', remove_first_level_duplicates=False)
        self.assertEqual(results_1[('target', 'Count')].sum(),  data.shape[0])
        hv.assert_is_close(results_1[('target', 'Count Perc')].sum(), 1)

        results = test_count_groups(group_1='target', remove_first_level_duplicates=True)
        assert_dataframes_match([results, results_1])

        results_2 = test_count_groups(group_1='checking_status', remove_first_level_duplicates=False)
        self.assertEqual(results_2[('checking_status', 'Count')].sum(), data.shape[0])
        hv.assert_is_close(results_2[('checking_status', 'Count Perc')].sum(), 1)

        results = test_count_groups(group_1='checking_status', remove_first_level_duplicates=True)
        assert_dataframes_match([results, results_2])

        results_3 = test_count_groups(
            group_1='target',
            group_2='checking_status',
            remove_first_level_duplicates=False
        )
        assert_dataframes_match([results_1,
                                 results_3[[('target', 'target'),
                                            ('target', 'Count'),
                                            ('target', 'Count Perc')]].drop_duplicates()])

        self.assertEqual(results_3[('checking_status', 'Count')].sum(), data.shape[0])

        group_2_count_sum = results_3.groupby(results_3[('target', 'target')]). \
            agg({('checking_status', 'Count'): sum})

        self.assertTrue((results_1[('target', 'Count')].values ==   # noqa
                         group_2_count_sum[('checking_status', 'Count')].values).all())
        self.assertEqual(len(group_2_count_sum), 3)

        group_2_count_perc_sum = results_3.groupby(results_3[('target', 'target')]).\
            agg({('checking_status', 'Count Perc'): sum})
        self.assertEqual(len(group_2_count_perc_sum), 3)
        self.assertTrue((group_2_count_perc_sum[('checking_status', 'Count Perc')] == 1).all())

        results = test_count_groups(
            group_1='target',
            group_2='checking_status',
            remove_first_level_duplicates=True
        )
        assert_dataframes_match([results_3[[('target', 'target'),
                                            ('target', 'Count'),
                                            ('target', 'Count Perc')]].astype(object).drop_duplicates(),
                                 results[[('target', 'target'),
                                          ('target', 'Count'),
                                          ('target', 'Count Perc')]].dropna().astype(object)])
        assert_dataframes_match([results_3[[('checking_status', 'checking_status'),
                                            ('checking_status', 'Count'),
                                            ('checking_status', 'Count Perc')]],
                                 results[[('checking_status', 'checking_status'),
                                          ('checking_status', 'Count'),
                                          ('checking_status', 'Count Perc')]]])

        results = test_count_groups(
            group_1='target',
            group_sum='credit_amount',
            remove_first_level_duplicates=False
        )
        assert_dataframes_match([results[[('target', 'target'),
                                          ('target', 'Count'),
                                          ('target', 'Count Perc')]].astype(object),
                                 results_1.astype(object)])
        self.assertEqual(results[('target', 'Sum')].sum(), data['credit_amount'].sum())
        self.assertEqual(results[('target', 'Sum Perc')].sum(), 1)

        results_x = test_count_groups(
            group_1='target',
            group_sum='credit_amount',
            remove_first_level_duplicates=True
        )
        assert_dataframes_match([results_x, results])

        results = test_count_groups(
            group_1='checking_status',
            group_sum='credit_amount',
            remove_first_level_duplicates=False
        )
        assert_dataframes_match([results[[('checking_status', 'checking_status'),
                                          ('checking_status', 'Count'),
                                          ('checking_status', 'Count Perc')]].astype(object),
                                 results_2.astype(object)])
        self.assertEqual(results[('checking_status', 'Sum')].sum(), data['credit_amount'].sum())
        self.assertEqual(results[('checking_status', 'Sum Perc')].sum(), 1)

        results_y = test_count_groups(
            group_1='checking_status',
            group_sum='credit_amount',
            remove_first_level_duplicates=True
        )
        assert_dataframes_match([results_y, results])

        results = test_count_groups(
            group_1='target',
            group_2='checking_status',
            group_sum='credit_amount',
            remove_first_level_duplicates=False
        )
        assert_dataframes_match([results_3.astype(object),
                                 results[[
                                     ('target', 'target'),
                                     ('target', 'Count'),
                                     ('target', 'Count Perc'),
                                     ('checking_status', 'checking_status'),
                                     ('checking_status', 'Count'),
                                     ('checking_status', 'Count Perc')
                                 ]].astype(object)])
        group_2_sum_perc_sum = results.groupby(results[('target', 'target')]). \
            agg({('checking_status', 'Sum Perc'): sum})
        self.assertEqual(len(group_2_sum_perc_sum), 3)
        self.assertTrue((group_2_sum_perc_sum[('checking_status', 'Sum Perc')].values == 1).all())

        results = test_count_groups(
            group_1='target',
            group_2='checking_status',
            group_sum='credit_amount',
            remove_first_level_duplicates=True
        )
        self.assertEqual(results[('checking_status', 'Sum')].sum(), data['credit_amount'].sum())
        self.assertEqual(results[('checking_status', 'Sum Perc')].sum(), 3)

        results = hp.count_groups(
            dataframe=data,
            group_1='target',
            group_2='checking_status',
            group_sum='credit_amount',
            return_style=True
        )
        with open(get_test_path('pandas/count_groups.html'), 'w') as file:
            file.write(clean_formatted_dataframe(results.to_html()))

        # test all of group 1 missing group_sum
        data.loc[data['target'] == True, 'credit_amount'] = np.nan  # noqa
        results = hp.count_groups(
            dataframe=data,
            group_1='target',
            group_2='checking_status',
            group_sum='credit_amount',
            return_style=False
        )
        target_counts = results[[('target', 'target'),
                                 ('target', 'Count'),
                                 ('target', 'Count Perc')]].dropna()
        assert_dataframes_match([target_counts.astype(object), results_1.astype(object)])
        hv.assert_is_close(results[('target', 'Sum')].sum(), data['credit_amount'].sum())
        hv.assert_is_close(results[('target', 'Sum Perc')].sum(), 1)
        self.assertTrue((results.loc[results[('target', 'target')] == True, ('target', 'Sum Perc')] == 0).all())  # noqa

        results = hp.count_groups(
            dataframe=data,
            group_1='target',
            group_2='checking_status',
            group_sum='credit_amount',
            return_style=True
        )
        with open(get_test_path('pandas/count_groups_group_1_na_sum.html'), 'w') as file:
            file.write(clean_formatted_dataframe(results.to_html()))


if __name__ == '__main__':
    unittest.main()



"""A collection of functions that assist in validation/comparison of data and conditions.
"""
from collections.abc import Sized
from typing import List, Union, Callable, Type, Iterable

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_categorical

from helpsk.exceptions import HelpskAssertionError, HelpskParamTypeError, HelpskParamValueError
from helpsk.utility import suppress_warnings


def any_none_nan(values: Union[List, np.ndarray, pd.Series, pd.DataFrame, object]) -> bool:
    """Can be used with a single value or a collection of values. Returns `True` if any item in `values` are
    `None`, `np.Nan`, `pd.NA`, `pd.NaT` or if the length of `values` is `0`.

    Args:
        values:
            A collection of values to check.

    Returns:
        bool - True if any item in `values` are None/np.NaN
    """
    if values is None or values is np.NaN or values is pd.NA or values is pd.NaT:
        return True

    if isinstance(values, Sized) and not isinstance(values, str) and len(values) == 0:
        return True

    if isinstance(values, pd.Series):
        return values.isnull().any() or values.isna().any()

    if isinstance(values, pd.DataFrame):
        return values.isnull().any().any() or values.isna().any().any()

    if isinstance(values, Iterable) and not isinstance(values, str):
        if len(values) == 0:
            return True

        return any((any_none_nan(x) for x in values))

    try:
        if not isinstance(values, str) and None in values:
            return True
    except Exception:
        pass

    try:
        if np.isnan(values).any():
            return True
    except TypeError:
        return False

    return False


def assert_not_none_nan(values: Union[List, np.ndarray, pd.Series, pd.DataFrame, object]) -> None:
    """Raises an HelpskAssertionError if any item in `values` are `None`, `np.Nan`, or if the length of
    `values` is `0`.

    For numeric types only.

    Args:
        values:
            A collection of values to check.
    """
    assert_false(any_none_nan(values), message='None/NaN Values Found')


def any_missing(values: Union[List, pd.Series, pd.DataFrame, object]) -> bool:
    """Same as `any_none_nan` but checks for empty strings

    Args:
        values:
            A collection of values to check.

    Returns:
        bool - True if any item in `values` are None/np.NaN/''
    """
    if any_none_nan(values):
        return True

    if isinstance(values, pd.Series):
        return values.isin(['']).any()  # noqa

    if isinstance(values, pd.DataFrame):
        return values.isin(['']).any().any()  # noqa

    if isinstance(values, str) and values.strip() == '':
        return True

    if isinstance(values, Iterable) and '' in values:
        return True

    return False


def assert_not_any_missing(values: Union[List, pd.Series, pd.DataFrame, object]) -> None:
    """Raises an HelpskAssertionError if any item in `values` are `None`, `np.Nan`, an empty string (i.e. '')
    or if the length of `values` is `0`.

    Args:
        values:
            A collection of values to check.
    """
    assert_false(any_missing(values), message='Missing Values Found')


def any_duplicated(values: Union[List, np.ndarray, pd.Series]) -> bool:
    """Returns `True` if any items in `values` are duplicated.

    Args:
        values: list, np.ndarray, pd.Series
            A collection of values to check.

    Returns:
        bool
    """
    return len(values) != len(set(values))


def assert_not_duplicated(values: Union[List, np.ndarray, pd.Series]) -> None:
    """Raises an HelpskAssertionError if any items in `values` are duplicated.

    Args:
        values: list, np.ndarray, pd.Series
            A collection of values to check.
    """
    assert_false(any_duplicated(values), message='Duplicate Values Found')


def assert_all(values: Union[List, np.ndarray, pd.Series, pd.DataFrame]) -> None:
    """Raises an `HelpskAssertionError` unless all items in `values` are `True`

    Args:
        values:
            A collection of values to check.
    """
    if isinstance(values, pd.Series):
        if not values.all():  # noqa
            raise HelpskAssertionError('Not All True')
    elif isinstance(values, pd.DataFrame):
        if not values.all().all():  # noqa
            raise HelpskAssertionError('Not All True')
    else:
        if not all(values):
            raise HelpskAssertionError('Not All True')


def assert_not_any(values: Union[List, np.ndarray, pd.Series, pd.DataFrame]) -> None:
    """Raises an `HelpskAssertionError` if any items in `values` are `True`

    Args:
        values:
            A collection of values to check.
    """
    if isinstance(values, pd.Series):
        assert_false(values.any(), message='Found True')  # noqa

    elif isinstance(values, pd.DataFrame):
        assert_false(values.any().any(), message='Found True')  # noqa

    else:
        assert_false(any(values), message='Found True')


def assert_true(condition: bool, message: str = 'Condition Not True') -> None:
    """Raises an HelpskAssertionError if `condition` is not True

    Args:
        condition:
            Something that evaluates to True/False
        message:
            Message passed to the HelpskAssertionError
    """
    if not isinstance(condition, (bool, np.bool_)):
        raise HelpskParamTypeError('condition should be boolean')

    if not condition:
        raise HelpskAssertionError(message)


def assert_false(condition: bool, message: str = 'Condition True') -> None:
    """Raises an HelpskAssertionError if `condition` is not False

    Args:
        condition: bool
            Something that evaluates to True/False
        message:
            Message passed to the HelpskAssertionError
    """
    if not isinstance(condition, (bool, np.bool_)):
        raise HelpskParamTypeError('condition should be boolean')

    if condition:
        raise HelpskAssertionError(message)


def iterables_are_equal(iterable_a: Iterable, iterable_b: Iterable) -> bool:
    """Compares the equality of the values of two iterables.

    This function will generally give the same result as list equality (e.g. `[x, y, z] == [x, y, z]`)
    However, in some strange scenarios, `==` will return `False` where it doesn't seem like it should

    For example:

    ```
    temp = pd.DataFrame({'col_a': [np.nan, 1.0]})
    temp.col_a.tolist() == [np.nan, 1.0]  # returns False. Why??
    iterables_are_equal(temp.col_a.tolist(), [np.nan, 1])  # returns True
    [np.nan, 1.0] == [np.nan, 1.0]  # returns True


    Also, when comparing a series with an ordered Categorical when the values are the same,
    pd.Series.equals() will return False if the categories have different order. But we only care if the
    values are the same, so this function will return True.
    ```

    Args:
        iterable_a:
            an iterable to equate to iterable_b
        iterable_b:
            an iterable to equate to iterable_a

    Returns:
        True if iterable_a is equal to iterable_b
    """
    # seems to be confusion and inconsistencies across stack overflow on how to properly check for category
    # so this might be overkill but not exactly sure
    # def is_categorical(series):
    #     if isinstance(series, (pd.Categorical, pd.CategoricalDtype)):
    #         return True
    #     if isinstance(series, pd.Series):
    #         return series.dtype.name == 'category'
    #     return False

    with suppress_warnings():
        # if either list-like structure is categorical, then we need to convert both to unordered categorical
        if is_categorical(iterable_a) or is_categorical(iterable_b):
            iterable_a = pd.Categorical(iterable_a, ordered=False)
            iterable_b = pd.Categorical(iterable_b, ordered=False)
        else:
            iterable_a = pd.Series(iterable_a)
            iterable_b = pd.Series(iterable_b)

        return iterable_a.equals(iterable_b)


def dataframes_match(dataframes: List[pd.DataFrame],
                     float_tolerance: int = 6,
                     ignore_indexes: bool = True,
                     ignore_column_names: bool = True) -> bool:
    """
    Because floating point numbers are difficult to accurate represent, when comparing multiple DataFrames,
    this function first rounds any numeric columns to the number of decimal points indicated
    `float_tolerance`.

    Args:
        dataframes:
            Two or more dataframes to compare against each other and test for equality

        float_tolerance:
            numeric columns will be rounded to the number of digits to the right of the decimal specified by
            this parameter.

        ignore_indexes:
            if True, the indexes of each DataFrame will be ignored for considering equality

        ignore_column_names:
            if True, the column names of each DataFrame will be ignored for considering equality

    Returns:
        Returns True if the dataframes match based on the conditions explained above, otherwise returns False
    """
    if not isinstance(dataframes, list):
        raise HelpskParamTypeError("Expected list of pd.DataFrame's.")

    if not len(dataframes) >= 2:
        raise HelpskParamValueError("Expected 2 or more pd.DataFrame's in list.")

    first_dataframe = dataframes[0].round(float_tolerance)

    def first_dataframe_equals_other(other_dataframe):
        if first_dataframe.shape != other_dataframe.shape:
            return False

        if ignore_indexes or ignore_column_names:
            # if either of these are True, then we are going to change the index and/or columns, but
            # python is pass-by-reference so we don't want to change the original DataFrame object.
            other_dataframe = other_dataframe.copy()

        if ignore_indexes:
            other_dataframe.index = first_dataframe.index

        if ignore_column_names:
            other_dataframe.columns = first_dataframe.columns

        return first_dataframe.equals(other_dataframe.round(float_tolerance))

    # compare the first dataframe to the rest of the dataframes, after rounding each to the tolerance, and
    # performing other modifications
    # check if all results are True
    return all(first_dataframe_equals_other(x) for x in dataframes[1:])


def assert_dataframes_match(dataframes: List[pd.DataFrame],
                            float_tolerance: int = 6,
                            ignore_indexes: bool = True,
                            ignore_column_names: bool = True,
                            message: str = 'Dataframes do not match') -> None:
    """
    Raises an assertion error if dataframes don't match.

    Args:
        dataframes:
            Two or more dataframes to compare against each other and test for equality

        float_tolerance:
            numeric columns will be rounded to the number of digits to the right of the decimal specified by
            this parameter.

        ignore_indexes:
            if True, the indexes of each DataFrame will be ignored for considering equality

        ignore_column_names:
            if True, the column names of each DataFrame will be ignored for considering equality

        message:
            message to pass to HelpskAssertionError
    """
    if not dataframes_match(dataframes=dataframes,
                            float_tolerance=float_tolerance,
                            ignore_indexes=ignore_indexes,
                            ignore_column_names=ignore_column_names):
        raise HelpskAssertionError(message)


def is_close(value_a: float, value_b: float, tolerance: float = 0.000001) -> bool:
    """Tests whether or not value_a and value_b are "close" (i.e. within the `tolerance` after subtracting)

    Args:
        value_a:
            numeric value to test
        value_b:
            numeric value to test
        tolerance:
            the maximum difference (absolute value) allowed between value_a and value_b
    Returns:
          True if values are within specified tolerance
    """
    return abs(value_a - value_b) <= tolerance


def assert_is_close(value_a: float, value_b: float, tolerance: float = 0.000001):
    """Raises an assert error if value_a and value_b are not "close" (see documentation of `is_close()`
    function).

    Args:
          value_a:
              numeric value to test
          value_b:
              numeric value to test
          tolerance:
              number of digits to round to
    """
    if not is_close(value_a=value_a, value_b=value_b, tolerance=tolerance):
        raise HelpskAssertionError(f"`{value_a}` and `{value_b}` are not within a tolerance of `{tolerance}`")


def raises_exception(function: Callable, exception_type: Type = None) -> bool:
    """Returns True if `function` raises an Exception; returns False if `function` runs without raising an
    Exception.
    Args:
        function:
            the function which does or does not raise an exception.
        exception_type:
            if `exception_type` is provided, `raises_exception` returns true only if the `function` argument
            raises an Exception **and** the exception has type of `exception_type`.
    """
    try:
        function()
        return False
    except Exception as exception:
        if exception_type:
            return isinstance(exception, exception_type)

        return True



import unittest

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

from helpsk.utility import suppress_warnings
from tests.helpers import check_plot, get_test_path


def plot_colors(colors, title, sort_colors=True):
    """
    Code from:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    """
    cell_width = 212
    cell_height = 22
    swatch_width = 48
    margin = 12
    top_margin = 40

    # Sort colors by hue, saturation, value and name.
    if sort_colors is True:
        by_hsv = sorted((tuple(mcolors.rgb_to_hsv(mcolors.to_rgb(color))),
                         name)
                        for name, color in colors.items())
        names = [name for hsv, name in by_hsv]
    else:
        names = list(colors)

    nrows = 7
    # n = len(names)
    # ncols = n // nrows + int(n % nrows > 0)
    # ncols = 4 - empty_cols
    # nrows = n // ncols + int(n % ncols > 0)

    width = cell_width * 4 + 2 * margin
    height = cell_height * nrows + margin + top_margin
    dpi = 72

    fig, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
    fig.subplots_adjust(margin/width, margin/height,
                        (width-margin)/width, (height-top_margin)/height)
    ax.set_xlim(0, cell_width * 4)
    ax.set_ylim(cell_height * (nrows-0.5), -cell_height/2.)
    ax.yaxis.set_visible(False)
    ax.xaxis.set_visible(False)
    ax.set_axis_off()
    ax.set_title(title, fontsize=24, loc="left", pad=10)

    for i, name in enumerate(names):
        row = i % nrows
        col = i // nrows
        y = row * cell_height

        swatch_start_x = cell_width * col
        text_pos_x = cell_width * col + swatch_width + 7

        ax.text(text_pos_x, y, name, fontsize=14,
                horizontalalignment='left',
                verticalalignment='center')

        with suppress_warnings():
            ax.add_patch(
                Rectangle(xy=(swatch_start_x, y-9), width=swatch_width,
                          height=18, color=colors[name], edgecolor='0.7')
            )

    return fig


class TestColors(unittest.TestCase):

    def test_plot_colors(self):
        from helpsk.color import Colors

        color_names = [e.name for e in Colors]
        color_values = [e.value for e in Colors]

        check_plot(file_name=get_test_path('color/colors.png'),
                   plot_function=lambda: plot_colors(dict(zip(color_names, color_values)),
                                                     title="Colors Enum", sort_colors=False))


if __name__ == '__main__':
    unittest.main()



import logging.config
import time
import unittest

from helpsk.logging import Timer, log_timer, log_function_call
from tests.helpers import get_test_path

logging.config.fileConfig(get_test_path("logging/test_logging.conf"),
                          defaults={'logfilename': get_test_path("logging/log.log")},
                          disable_existing_loggers=False)


class TestLogging(unittest.TestCase):

    def test__timer(self):
        print('\n')

        with Timer("testing timer"):
            time.sleep(0.05)

        with Timer("testing another timer") as timer:
            time.sleep(0.1)

        self.assertIsNotNone(timer._interval)

    def test__timer_decorator(self):
        print('\n')

        @log_timer
        def my_function_1(param_1, param_2):
            time.sleep(0.1)
            return param_1, param_2

        @log_timer
        def my_function_2():
            time.sleep(0.1)

        # make sure we call the @wraps function
        self.assertEqual(my_function_1.__name__, 'my_function_1')
        self.assertEqual(my_function_2.__name__, 'my_function_2')

        my_function_2()

        a, b = my_function_1(1, param_2=2)
        self.assertEqual(a, 1)
        self.assertEqual(b, 2)
        a, b = my_function_1(param_1='value-1', param_2='value-2')
        self.assertEqual(a, 'value-1')
        self.assertEqual(b, 'value-2')

    def test__log_function(self):
        @log_function_call
        def my_function_1(param_1, param_2):
            return param_1, param_2

        @log_function_call
        def my_function_2():
            return 'value'

        # make sure we call the @wraps function
        self.assertEqual(my_function_1.__name__, 'my_function_1')
        self.assertEqual(my_function_2.__name__, 'my_function_2')

        value = my_function_2()
        self.assertEqual(value, 'value')

        a, b = my_function_1(1, param_2=2)
        self.assertEqual(a, 1)
        self.assertEqual(b, 2)
        a, b = my_function_1(param_1='value-1', param_2='value-2')
        self.assertEqual(a, 'value-1')
        self.assertEqual(b, 'value-2')



import unittest
from os import remove
import os.path
import time
import pandas as pd

from helpsk.utility import to_pickle, read_pickle, open_yaml, Timer, is_debugging, dataframe_to_pickle, \
    dataframe_to_csv, object_to_pickle
from tests.helpers import get_test_path


# noinspection PyMethodMayBeStatic
class TestValidation(unittest.TestCase):

    def test_pickle(self):
        obj = "this is an object"
        path = get_test_path('pickled_object.pkl')
        self.assertFalse(os.path.isfile(path))
        to_pickle(obj=obj, path=path)
        self.assertTrue(os.path.isfile(path))
        unpickled_object = read_pickle(path=path)
        self.assertEqual(obj, unpickled_object)
        remove(path)
        self.assertFalse(os.path.isfile(path))

    def test__open_yaml(self):
        config_yaml = open_yaml(get_test_path('utility/config.yaml'))
        self.assertIsInstance(config_yaml, dict)
        self.assertIsNotNone(config_yaml['CONFIG']['DIRECTORY'])

    def test_is_debugging(self):
        # not sure how to test is_debugging is true, except manually
        self.assertFalse(is_debugging())

    def test__timer(self):
        print('\n')

        with Timer("testing timer"):
            time.sleep(0.05)

        with Timer("testing another timer") as timer:
            time.sleep(0.1)

        self.assertIsNotNone(timer._interval)

    def test__dataframe_to_pickle(self):
        df = pd.DataFrame(['test'])
        output_directory = 'temp_test'
        file_name = 'test.pkl'
        self.assertFalse(os.path.isdir(output_directory))
        file_path = dataframe_to_pickle(df=df, output_directory=output_directory, file_name=file_name)
        self.assertTrue(os.path.isfile(file_path))
        self.assertEqual(file_path, 'temp_test/test.pkl')
        df_unpickled = pd.read_pickle(file_path)
        self.assertTrue((df_unpickled.iloc[0] == df.iloc[0]).all())  # noqa
        os.remove(file_path)
        os.removedirs(output_directory)
        self.assertFalse(os.path.isdir(output_directory))

    def test__dataframe_to_csv(self):
        df = pd.DataFrame(['test'])
        output_directory = 'temp_test'
        file_name = 'test.csv'
        self.assertFalse(os.path.isdir(output_directory))
        file_path = dataframe_to_csv(df=df, output_directory=output_directory, file_name=file_name)
        self.assertTrue(os.path.isfile(file_path))
        self.assertEqual(file_path, 'temp_test/test.csv')
        df_unpickled = pd.read_csv(file_path)
        self.assertEqual(df_unpickled.loc[0, '0'], df.iloc[0, 0])
        os.remove(file_path)
        os.removedirs(output_directory)
        self.assertFalse(os.path.isdir(output_directory))

    def test__object_to_pickle(self):
        df = pd.DataFrame(['test'])
        output_directory = 'temp_test'
        file_name = 'test.pkl'
        self.assertFalse(os.path.isdir(output_directory))
        file_path = object_to_pickle(obj=df, output_directory=output_directory, file_name=file_name)
        self.assertTrue(os.path.isfile(file_path))
        self.assertEqual(file_path, 'temp_test/test.pkl')
        df_unpickled = pd.read_pickle(file_path)
        self.assertTrue((df_unpickled.iloc[0] == df.iloc[0]).all())  # noqa
        os.remove(file_path)
        os.removedirs(output_directory)
        self.assertFalse(os.path.isdir(output_directory))


if __name__ == '__main__':
    unittest.main()



"""
init file
"""
# from .example_class import ExampleClass
# https://stackoverflow.com/questions/8899198/module-has-no-attribute
import helpsk.color
import helpsk.database
import helpsk.database_base
import helpsk.date
import helpsk.exceptions
import helpsk.pandas
import helpsk.pandas_style
import helpsk.plot
import helpsk.sklearn_eval
import helpsk.sklearn_pipeline
import helpsk.string
import helpsk.utility
import helpsk.validation  # noqa



import datetime
import unittest
# noinspection PyMethodMayBeStatic
from enum import unique, Enum, auto

import numpy as np
import pandas as pd

import helpsk.plot as hplot
import helpsk.validation as hval
from tests.helpers import get_data_credit, check_plot, get_test_path


@unique
class TestEnum(Enum):
    VALUE_A = auto()
    VALUE_B = auto()


class TestPlot(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.credit_data = get_data_credit()
        sample_data = pd.DataFrame({'col_a': [np.nan, 2, 3, 4],
                                    'col_b': [np.nan, 'b', 'd', 'd'],
                                    'col_c': pd.date_range('2021-01-01', '2021-01-04'),
                                    'col_d': [None, np.nan,
                                              datetime.date(2021, 4, 2), datetime.date(2021, 4, 2)],
                                    'col_e': np.nan,
                                    'col_f': [1.0, 2.0, 3.0, 4.0],
                                    'col_h': [np.nan, TestEnum.VALUE_A, TestEnum.VALUE_B, TestEnum.VALUE_B],
                                    'col_i': [None, np.nan, datetime.datetime(2021, 4, 2, 0, 0, 0),
                                              datetime.datetime(2021, 4, 2, 0, 0, 0)],
                                    })
        sample_data.loc[0, 'col_c'] = np.nan
        sample_data['col_g'] = sample_data['col_b'].astype('category')
        cls.sample_data = sample_data

    def test_plot_value_frequency(self):
        test_series = self.credit_data['purpose'].copy()
        test_series[0:10] = np.nan

        from helpsk.pandas import value_frequency
        value_frequency(series=test_series)

        check_plot(file_name=get_test_path('plot/test_plot_value_frequency__sort_true.png'),
                   plot_function=lambda: hplot.plot_value_frequency(series=test_series,
                                                                    sort_by_frequency=True))

        check_plot(file_name=get_test_path('plot/test_plot_value_frequency__sort_false.png'),
                   plot_function=lambda: hplot.plot_value_frequency(series=test_series,
                                                                    sort_by_frequency=False))

    def test_plot_correlation_heatmap(self):
        test_series = self.credit_data.copy()
        test_series.loc[0:10, 'credit_amount'] = np.nan
        test_series.loc[0:10, 'num_dependents'] = np.nan

        check_plot(file_name=get_test_path('plot/test_plot_correlation_heatmap__default.png'),
                   plot_function=lambda: hplot.plot_correlation_heatmap(test_series))

        check_plot(
            file_name=get_test_path('plot/test_plot_correlation_heatmap__params.png'),
            plot_function=lambda: hplot.plot_correlation_heatmap(
                test_series,
                threshold=0.2,
                title='Credit Data: Correlations Above 0.2',
                figure_size=(6, 6),
                round_by=3,
                features_to_highlight=['credit_amount', 'num_dependents']
            )
        )

    def test_plot_dodged_barchart(self):
        test_series = self.credit_data.copy()
        test_series.loc[0:10, 'credit_history'] = np.nan
        test_series.loc[5:15, 'own_telephone'] = np.nan

        original = test_series.copy()

        check_plot(file_name=get_test_path('plot/plot_dodged_barchart__default.png'),
                   plot_function=lambda: hplot.plot_dodged_barchart(dataframe=test_series,
                                                                    outer_column='own_telephone',
                                                                    inner_column='credit_history'))

        self.assertTrue(hval.dataframes_match([original, test_series]))

    def test_plot_dodged_barchart__boolean(self):
        """There is a bug where using a boolean series fails"""
        test_series = self.credit_data.copy()
        test_series['all_paid'] = test_series['credit_history'] == 'all paid'
        test_series.loc[0:10, 'all_paid'] = np.nan

        test_series['own_telephone'] = test_series['own_telephone'] == 'yes'
        test_series.loc[5:15, 'own_telephone'] = None

        original = test_series.copy()

        check_plot(file_name=get_test_path('plot/plot_dodged_barchart__booleans.png'),
                   plot_function=lambda: hplot.plot_dodged_barchart(dataframe=test_series,
                                                                    outer_column='own_telephone',
                                                                    inner_column='all_paid'))

        self.assertTrue(hval.dataframes_match([original, test_series]))

    def test_plot_histogram_with_categorical(self):
        test_series = self.credit_data.copy()
        test_series.loc[0:10, 'credit_history'] = np.nan
        test_series.loc[5:15, 'credit_amount'] = np.nan

        original = test_series.copy()

        check_plot(
            file_name=get_test_path('plot/test_plot_histogram_with_categorical__default.png'),
            plot_function=lambda: hplot.plot_histogram_with_categorical(
                dataframe=test_series,
                numeric_column='credit_amount',
                categorical_column='credit_history'
            )
        )

        self.assertTrue(hval.dataframes_match([original, test_series]))

    def test_plot_histogram_with_categorical_bug(self):
        # bug found when no missing values are found in target because we added a <missing> category but
        # we get an error when calling reorder_categories with extra categories
        credit_data = self.credit_data.copy()
        hplot.plot_histogram_with_categorical(dataframe=credit_data,
                                              numeric_column='credit_amount',
                                              categorical_column='target')


if __name__ == '__main__':
    unittest.main()



import os
import re
from importlib import reload
from typing import Callable, Union
from unittest import TestCase

import pandas as pd
from matplotlib import pyplot as plt

from helpsk.pandas import print_dataframe
from helpsk.utility import redirect_stdout_to_file


def get_test_path(file) -> str:
    """Returns the path to /tests folder, adjusting for the difference in the current working directory when
    debugging vs not debugging.
    """
    path = os.getcwd()
    return os.path.join(path, 'tests/test_files', file)


def subtests_expected_vs_actual(test_case: TestCase,
                                expected_values: list,
                                actual_values: list,
                                **kwargs) -> None:
    """ runs a TestCase.subTest() for each item in expected_values and actual_values and tests that the
    expected_values are equal to the actual_values via TestCase.assertEqual()

    Example (ran within `test_example_file.py -> TestExampleClass -> test_example_function()`):

        ```
        expected_values = [1, 2, 3, 4]
        actual_values = [1, 0, 3, -1]

        subTests_actual_expected(test_case=self, expected_values=expected_values, actual_values=actual_values,
                                 my_param1=True, my_param2='value')

        ```

        The output will be:

        ```
        FAIL: test_example_function(test_example_file.TestExampleClass)
            (index=1, expected=2, actual=0, my_param1=True, my_param2='value')
        AssertionError: 2 != 0

        FAIL: test_example_function(test_example_file.TestExampleClass)
            (index=3, expected=4, actual=-1, my_param1=True, my_param2='value')
        AssertionError: 4 != -1
        ```

    Args:
        test_case:
            the current TestCase object (i.e. pass `self` from within a test function)
        expected_values:
            a list of expected values
        actual_values:
            a list of actual values
        **kwargs:
            a variable list of param/value items that will get passed to TestCase.subTest and printed out for
            failing sub-tests.
    """
    assert len(expected_values) == len(actual_values)

    for index, (expected, actual) in enumerate(zip(expected_values, actual_values)):
        with test_case.subTest(index=index, expected=expected, actual=actual, **kwargs):
            test_case.assertEqual(expected, actual)


def check_plot(file_name: str, plot_function: Callable, set_size_w_h: Union[tuple, None] = (10, 6)):
    reload(plt)  # necessary because matplotlib throws strange errors about alpha values

    def clear():
        plt.gcf().clear()
        plt.cla()
        plt.clf()
        plt.close()

    clear()
    if os.path.isfile(file_name):
        os.remove(file_name)
    assert os.path.isfile(file_name) is False
    plot_function()
    if set_size_w_h is not None:
        fig = plt.gcf()
        fig.set_size_inches(set_size_w_h[0], set_size_w_h[1])
    plt.savefig(file_name)
    clear()
    assert os.path.isfile(file_name)


def clean_formatted_dataframe(rendered):
    """`This dataframe changes the random code generated when saving formatted dataframes
    (i.e df.style.to_html()). This is necessary because it each time unit tests are ran, the html changes,
    and it is difficult to know if the change/diff occurred because of the random code, or because something
    actually changed.

    For example, it changes the `1ef9B` in

    ```
    '<style type="text/css">\n#T_1ef9B_row0_col0, #T_1ef9B_row0_col1, ...
    ```

    to `99999`:

    ```
    '<style type="text/css">\n#T_99999_row0_col0, #T_99999_row0_col1, ...
    ```
    `"""
    temp = rendered.replace('\n', '')
    if temp.startswith('<style type="text/css"></style>'):
        code = re.sub('">  <.*', '', temp).replace('<style type="text/css"></style><table id="T', '')
    else:
        code = re.sub('_row.*', '_', temp).replace('<style type="text/css">#T', '')

    if code.endswith('_'):
        code = code[:-1]

    return rendered.replace(code, '_99999')


def get_data_credit() -> pd.DataFrame:
    # https://www.openml.org/d/31
    # save locally in case dataset changes or is removed
    # from sklearn.datasets import fetch_openml
    # credit_g = fetch_openml('credit-g', version=1)
    # data = credit_g['data']
    # data['target'] = credit_g['target']
    # data.to_pickle(get_test_path('../test_data/credit.pkl'))
    return pd.read_pickle(get_test_path('../test_data/credit.pkl'))


def get_data_titanic() -> pd.DataFrame:
    # save locally in case dataset changes or is removed
    # from sklearn.datasets import fetch_openml
    # titanic = fetch_openml('titanic', version=1)
    # data = titanic['data']
    # data['survived'] = titanic['target']
    # data.to_pickle(get_test_path('../test_data/titanic.pkl'))
    return pd.read_pickle(get_test_path('../test_data/titanic.pkl'))


def get_data_housing() -> pd.DataFrame:
    # https://www.openml.org/d/537
    # save locally in case dataset changes or is removed
    # from sklearn.datasets import fetch_openml
    # housing = fetch_openml('houses', version=1)
    # data = housing['data']
    # data['target'] = housing['target']
    # data.to_pickle(get_test_path('../test_data/housing.pkl'))
    return pd.read_pickle(get_test_path('../test_data/housing.pkl'))


def helper_test_dataframe(file_name, dataframe):
    if os.path.isfile(file_name):
        os.remove(file_name)

    pd.set_option('display.max_rows', None)

    with redirect_stdout_to_file(file_name):
        print_dataframe(dataframe)
    assert os.path.isfile(file_name)
